@@ -16,30 +16,32 @@
 ## -- 2021-12-19  1.1.3     DA       Replaced 'done' by 'success'
 ## -- 2021-12-21  1.1.4     DA       Class RobotHTM: renamed method reset() to _reset()
 ## -- 2021-12-21  1.1.5     MRD      Add Termination on Success
+## -- 2022-01-21  1.1.6     MRD      Add recommended cycle limit and add seed parameter
+## -- 2022-02-25  1.1.7     SY       Refactoring due to auto generated ID in class Dimension
+## -- 2022-11-09  1.1.8     DA       Refactoring due to changes on plot systematics
 ## -------------------------------------------------------------------------------------------------
 
 """
-Ver. 1.1.5 (2021-12-21)
+Ver. 1.1.8 (2022-11-09)
 
-This module provide an environment of a robot manipulator based on Homogeneous Matrix
+This module provides an environment of a robot manipulator based on Homogeneous Matrix
 """
+
 import torch
 import transformations
-import numpy as np
-import random
 from mlpro.rl.models import *
 
 
 
-
 ## -------------------------------------------------------------------------------------------------
 ## -------------------------------------------------------------------------------------------------
 class RobotArm3D:
     """
-    Auxilary class for the implementation of robotinhtm.
+    Auxiliary class for the implementation of robotinhtm.
     Generate the Kinematic of a pre-defined robot in Homogeneous Matrix.
     """
 
+## -------------------------------------------------------------------------------------------------
     def __init__(self):
         self.thetas = torch.Tensor([])
         self.joints = torch.Tensor([])
@@ -52,14 +54,15 @@ class RobotArm3D:
         self.HM = torch.Tensor([])
         self.HMeef = None
 
+
 ## -------------------------------------------------------------------------------------------------
     def add_link_joint(
-        self,
-        jointAxis=None,
-        lvector=None,
-        thetaInit=None,
-        adjustRot=None,
-        adjustTheta=None,
+            self,
+            jointAxis=None,
+            lvector=None,
+            thetaInit=None,
+            adjustRot=None,
+            adjustTheta=None,
     ):
         self.joints = torch.cat([self.joints, torch.Tensor([[0, 0, 0, 1]]).T], dim=1)
         self.jointAxis.append(jointAxis)
@@ -74,6 +77,7 @@ class RobotArm3D:
             self.thetas = torch.cat([self.thetas, thetaInit])
             self.num_joint = self.num_joint + 1
 
+
 ## -------------------------------------------------------------------------------------------------
     def get_transformation_matrix(self, theta, lvector, rotAxis, adjustRots=None, adjustThetas=None):
         transformationMatrix = torch.Tensor([])
@@ -231,58 +235,73 @@ class RobotArm3D:
     def update_theta(self, deltaTheta):
         self.thetas += deltaTheta.flatten()
 
-    ## -------------------------------------------------------------------------------------------------
+
+## -------------------------------------------------------------------------------------------------
     def convert_to_quaternion(self):
-         origin, xaxis, yaxis, zaxis = [0, 0, 0], [1, 0, 0], [0, 1, 0], [0, 0, 1]
-         hm_reshape = self.HM.reshape(self.joints.shape[1],4,4)
-         hm_length = hm_reshape.shape[0]
+        origin, xaxis, yaxis, zaxis = [0, 0, 0], [1, 0, 0], [0, 1, 0], [0, 0, 1]
+        hm_reshape = self.HM.reshape(self.joints.shape[1], 4, 4)
+        hm_length = hm_reshape.shape[0]
 
-         out = torch.Tensor([])
-         for x in range(hm_length):
-             _,_,rot,trans,_ = transformations.decompose_matrix(hm_reshape[x].numpy())
+        out = torch.Tensor([])
+        for x in range(hm_length):
+            _, _, rot, trans, _ = transformations.decompose_matrix(hm_reshape[x].numpy())
 
-             qx = transformations.quaternion_about_axis(rot[0], xaxis)
-             qy = transformations.quaternion_about_axis(rot[1], yaxis)
-             qz = transformations.quaternion_about_axis(rot[2], zaxis)
-             q = transformations.quaternion_multiply(qx, qy)
-             q = transformations.quaternion_multiply(q, qz)
+            qx = transformations.quaternion_about_axis(rot[0], xaxis)
+            qy = transformations.quaternion_about_axis(rot[1], yaxis)
+            qz = transformations.quaternion_about_axis(rot[2], zaxis)
+            q = transformations.quaternion_multiply(qx, qy)
+            q = transformations.quaternion_multiply(q, qz)
 
-             outs = torch.Tensor(trans)
-             outs = torch.cat([outs, torch.Tensor(q)])
-             out = torch.cat([out,outs])
-
-         return out    
+            outs = torch.Tensor(trans)
+            outs = torch.cat([outs, torch.Tensor(q)])
+            out = torch.cat([out, outs])
 
+        return out
 
 
 
 
 ## -------------------------------------------------------------------------------------------------
 ## -------------------------------------------------------------------------------------------------
-class RobotHTM(Environment):
+class RobotHTM (Environment):
     """
-    Custom environment for a arm robot represented as Homogeneous Matrix.
+    Custom environment for an arm robot represented as Homogeneous Matrix.
     The goal is to reach a certain point.
+
+    Parameters
+    ----------
+    p_num_joints : int
+        Number of joints. Default = 4.
+    p_reset_seed : bool
+        If True, random generator is reset. Default = True.
+    p_seed
+        Seeding value for the random generator. Default = None.
+    p_target_mode : str
+        Target mode. Possible values are "random" (default) or "fixed".
+    p_visualize : bool
+        Boolean switch for env/agent visualisation. Default = True.
+    p_logging
+        Log level (see constants of class Log). Default = Log.C_LOG_ALL.
     """
 
-    C_NAME = "RobotHTM"
-    C_REWARD_TYPE = Reward.C_TYPE_OVERALL
-    C_LATENCY = timedelta(0, 1, 0)
-    C_INFINITY = np.finfo(np.float32).max
+    C_NAME          = "RobotHTM"
+    C_REWARD_TYPE   = Reward.C_TYPE_OVERALL
+    C_LATENCY       = timedelta(0, 1, 0)
+    C_INFINITY      = np.finfo(np.float32).max
+    C_CYCLE_LIMIT   = 100
 
 ## -------------------------------------------------------------------------------------------------
-    def __init__(self, p_num_joints=4, p_target_mode="Random", p_logging=Log.C_LOG_ALL):
-        """
-        Parameters:
-            p_logging               Boolean switch for logging
-            p_num_joints            Number of joints
-            p_target_mode           Target Mode (Random or Fixed)
-        """
+    def __init__( self, 
+                  p_num_joints = 4, 
+                  p_reset_seed = True, 
+                  p_seed = None, 
+                  p_target_mode : str = "random", 
+                  p_visualize : bool = True,
+                  p_logging = Log.C_LOG_ALL ):
 
         self.RobotArm1 = RobotArm3D()
 
-        roboconf = {}
-        roboconf["Joints"] = []
+        joints = []
 
         jointType = []
         vectLinkLength = [[0, 0, 0], [0, 0, 0]]
@@ -300,12 +319,9 @@ class RobotHTM(Environment):
                 Joint_type=jointType[x],
                 Vector_link_length=vectorLink,
             )
-            roboconf["Joints"].append(joint)
+            joints.append(joint)
 
-        roboconf["Target_mode"] = p_target_mode
-        roboconf["Update_rate"] = 0.01
-
-        for robo in roboconf["Joints"]:
+        for robo in joints:
             self.RobotArm1.add_link_joint(
                 lvector=torch.Tensor(
                     [
@@ -322,16 +338,18 @@ class RobotHTM(Environment):
 
         self.RobotArm1.update_joint_coords()
         self.jointangles = self.RobotArm1.thetas
-        self.dt = roboconf["Update_rate"]
-        self.target_mode = roboconf["Target_mode"]
+        self.dt = 0.01
+        self.target_mode = p_target_mode
         self.target = None
         self.init_distance = None
         self.num_joint = self.RobotArm1.get_num_joint()
         self.reach = torch.norm(torch.Tensor([[0.0, 0.0, 0.0]]) - self.RobotArm1.joints[:3, [-1]].reshape(1, 3))
         self.last_distance = None
 
-        super().__init__(p_mode=Environment.C_MODE_SIM, p_logging=p_logging)
+        super().__init__(p_mode=Environment.C_MODE_SIM, p_visualize=p_visualize, p_logging=p_logging)
         self._state_space, self._action_space = self._setup_spaces()
+        self.set_random_seed(p_seed)
+        self._reset_seed = p_reset_seed
 
         self.reset()
 
@@ -351,17 +369,16 @@ class RobotHTM(Environment):
 
         # 1 Setup state space
         state_space = ESpace()
-        state_space.add_dim(Dimension(0, "Tx", "Targetx", "", "m", "m", p_boundaries=[-np.inf, np.inf]))
-        state_space.add_dim(Dimension(1, "Ty", "Targety", "", "m", "m", p_boundaries=[-np.inf, np.inf]))
-        state_space.add_dim(Dimension(2, "Tz", "Targetz", "", "m", "m", p_boundaries=[-np.inf, np.inf]))
-        state_space.add_dim(Dimension(3, "Px", "Targetx", "", "m", "m", p_boundaries=[-np.inf, np.inf]))
-        state_space.add_dim(Dimension(4, "Py", "Targety", "", "m", "m", p_boundaries=[-np.inf, np.inf]))
-        state_space.add_dim(Dimension(5, "Pz", "Targetz", "", "m", "m", p_boundaries=[-np.inf, np.inf]))
+        state_space.add_dim(Dimension("Tx", "Targetx", "", "m", "m", p_boundaries=[-np.inf, np.inf]))
+        state_space.add_dim(Dimension("Ty", "Targety", "", "m", "m", p_boundaries=[-np.inf, np.inf]))
+        state_space.add_dim(Dimension("Tz", "Targetz", "", "m", "m", p_boundaries=[-np.inf, np.inf]))
+        state_space.add_dim(Dimension("Px", "Targetx", "", "m", "m", p_boundaries=[-np.inf, np.inf]))
+        state_space.add_dim(Dimension("Py", "Targety", "", "m", "m", p_boundaries=[-np.inf, np.inf]))
+        state_space.add_dim(Dimension("Pz", "Targetz", "", "m", "m", p_boundaries=[-np.inf, np.inf]))
 
         for idx in range(self.num_joint):
             state_space.add_dim(
                 Dimension(
-                    idx + 6,
                     "J%i" % (idx),
                     "Joint%i" % (idx),
                     "",
@@ -376,7 +393,6 @@ class RobotHTM(Environment):
         for idx in range(self.num_joint):
             action_space.add_dim(
                 Dimension(
-                    idx,
                     "A%i" % (idx),
                     "AV%i" % (idx),
                     "",
@@ -413,7 +429,7 @@ class RobotHTM(Environment):
         return state
 
 
- ## -------------------------------------------------------------------------------------------------
+## -------------------------------------------------------------------------------------------------
     def _compute_success(self, p_state: State = None) -> bool:
         disterror = np.linalg.norm(np.array(p_state.get_values())[:3] - np.array(p_state.get_values())[3:6])
 
@@ -453,7 +469,8 @@ class RobotHTM(Environment):
 
 ## -------------------------------------------------------------------------------------------------
     def _reset(self, p_seed=None) -> None:
-        self.set_random_seed(p_seed)
+        if self._reset_seed:
+            self.set_random_seed(p_seed)
         theta = torch.zeros(self.RobotArm1.get_num_joint())
         self.RobotArm1.set_theta(theta)
         self.RobotArm1.update_joint_coords()
@@ -488,14 +505,4 @@ class RobotHTM(Environment):
         )
         obs = obs.cpu().flatten().tolist()
         self._state = State(self._state_space)
-        self._state.set_values(obs)
-
-
-## -------------------------------------------------------------------------------------------------
-    def init_plot(self, p_figure=None):
-        pass
-
-
-## -------------------------------------------------------------------------------------------------
-    def update_plot(self):
-        pass
+        self._state.set_values(obs)
\ No newline at end of file
