@@ -41,10 +41,17 @@
 ## -- 2022-07-27  1.4.1     DA       Introduction of root class Wrapper
 ## -- 2022-07-28  1.4.2     SY       Minor improvements: API documentation and logging
 ## -- 2022-08-15  1.4.3     DA       Correction of integration of class Wrapper
+## -- 2022-10-08  1.4.4     SY       Bug fixing and minor improvements: return of the reset function
+## -- 2022-11-01  1.4.5     DA       Refactoring
+## -- 2022-11-09  1.4.6     DA       Refactoring
+## -- 2022-11-29  1.4.7     DA       Refactoring
+## -- 2023-01-14  1.4.8     MRD      Separate reset function for gym, reset_old and reset_new
+## -- 2023-02-18  1.5.0     DA       Added minimum version 0.21.0
+## -- 2023-02-20  1.6.0     DA       Class WrEnvGym2MLPro: specific implementations for load(), _save()
 ## -------------------------------------------------------------------------------------------------
 
 """
-Ver. 1.4.3 (2022-08-15)
+Ver. 1.6.0 (2023-02-20)
 
 This module provides wrapper classes for OpenAI Gym environments.
 
@@ -53,9 +60,10 @@ See also: https://pypi.org/project/gym
 """
 
 import gym
+import dill as pkl
 from gym.core import Env
 from mlpro.wrappers.models import Wrapper
-from mlpro.rl.models import *
+from mlpro.rl import *
 
 
 
@@ -64,9 +72,9 @@ from mlpro.rl.models import *
 ## -------------------------------------------------------------------------------------------------
 class WrEnvGYM2MLPro(Wrapper, Environment):
     """
-    This class is a ready to use wrapper class for OpenAI Gym environments. 
-    Objects of this type can be treated as an environment object. Encapsulated 
-    gym environment must be compatible to class gym.Env.
+    This class is a ready to use wrapper class for OpenAI Gym environments. Objects of this type 
+    can be treated as an environment object. Encapsulated gym environment must be compatible to 
+    class gym.Env.
 
     Parameters
     ----------
@@ -76,25 +84,30 @@ class WrEnvGYM2MLPro(Wrapper, Environment):
         Optional external state space object that meets the state space of the Gym environment
     p_action_space : MSpace 
         Optional external action space object that meets the action space of the Gym environment
+    p_visualize : bool
+        Boolean switch for env/agent visualisation. Default = True.
     p_logging
         Log level (see constants of class Log). Default = Log.C_LOG_ALL.
     """
 
-
-    C_TYPE              = 'Wrapper OpenAI Gym -> MLPro'
+    C_TYPE              = 'Wrapper Gym2MLPro'
     C_WRAPPED_PACKAGE   = 'gym'
+    C_MINIMUM_VERSION   = '0.21.0'
+    C_PLOT_ACTIVE: bool = True
 
 ## -------------------------------------------------------------------------------------------------
     def __init__(self,
                  p_gym_env,  
                  p_state_space: MSpace = None,  
                  p_action_space: MSpace = None,  
+                 p_visualize:bool=True,
                  p_logging=Log.C_LOG_ALL):
 
-        self._gym_env = p_gym_env
-        self.C_NAME = 'Env "' + self._gym_env.spec.id + '"'
+        self._gym_env    = p_gym_env
+        self._gym_env_id = self._gym_env.env.spec.id
+        self.C_NAME      = '(' + self._gym_env_id + ')'
 
-        Environment.__init__(self, p_mode=Environment.C_MODE_SIM, p_latency=None, p_logging=p_logging)
+        Environment.__init__(self, p_mode=Environment.C_MODE_SIM, p_latency=None, p_visualize=p_visualize, p_logging=p_logging)
         Wrapper.__init__(self, p_logging=p_logging)
 
         if p_state_space is not None:
@@ -123,6 +136,43 @@ class WrEnvGYM2MLPro(Wrapper, Environment):
             pass
 
 
+## -------------------------------------------------------------------------------------------------
+    @staticmethod
+    def load(p_path, p_filename):
+        mlpro_env = pkl.load(open(p_path + os.sep + p_filename, 'rb'))
+        mlpro_env._gym_env = gym.make(mlpro_env._gym_env_id)
+        return mlpro_env
+
+
+## -------------------------------------------------------------------------------------------------
+    def _save(self, p_path, p_filename) -> bool:
+        """
+        The embedded Gym env itself can't be pickled due to it's dependencies on Pygame. That's why
+        the current env instance needs to be removed before pickling the object. After that a fresh
+        Gym env with same id is instantiated. Current inner state informations get lost.
+
+        See also: https://stackoverflow.com/questions/52336196/how-to-save-object-using-pygame-surfaces-to-file-using-pickle
+        """
+
+        # 1 Try to pickle/dill the entire environment...
+        try:
+            pkl.dump( obj=self, 
+                      file=open(p_path + os.sep + p_filename, "wb"),
+                      protocol=pkl.HIGHEST_PROTOCOL )
+            self.log(Log.C_LOG_TYPE_I, 'Env successfully stored')
+        except:
+            pass
+
+        # 2 Store the empty wrapper object without embedded env and re-initialize 
+        self._gym_env = None 
+        pkl.dump( obj=self, 
+                  file=open(p_path + os.sep + p_filename, "wb"),
+                  protocol=pkl.HIGHEST_PROTOCOL )
+        self._gym_env = gym.make(self._gym_env_id)
+        self.log(Log.C_LOG_TYPE_W, 'Wrapper stored without env. Env is recovered, but state was lost.')
+        return True
+
+
 ## -------------------------------------------------------------------------------------------------
     @staticmethod
     def recognize_space(p_gym_space) -> ESpace:
@@ -145,7 +195,7 @@ class WrEnvGYM2MLPro(Wrapper, Environment):
 
         if isinstance(p_gym_space, gym.spaces.Discrete):
             space.add_dim(
-                Dimension(p_name_short='0', p_base_set=Dimension.C_BASE_SET_Z, p_boundaries=[p_gym_space.n]))
+                Dimension(p_name_short='0', p_base_set=Dimension.C_BASE_SET_Z, p_boundaries=[0, int(p_gym_space.n-1)]))
         elif isinstance(p_gym_space, gym.spaces.Box):
             shape_dim = len(p_gym_space.shape)
             for i in range(shape_dim):
@@ -240,7 +290,6 @@ class WrEnvGYM2MLPro(Wrapper, Environment):
                     observation, reward_gym, done, info = self._gym_env.step(np.atleast_1d(action_gym))
         except:
             # For gym version below than 0.25 (This will be removed soon)
-            self.log(self.C_LOG_TYPE_W, 'Please upgrade your gym version to 0.25.0 or above. This behaviour will be removed in near future.')
             try:
                 observation, reward_gym, done, info = self._gym_env.step(action_gym)
             except:
@@ -339,21 +388,17 @@ class WrEnvGYM2MLPro(Wrapper, Environment):
 
 
 ## -------------------------------------------------------------------------------------------------
-    def init_plot(self, p_figure=None):
-        """
-        Plot initialization function, deployed by render functionality from OpenAI Gym.
-
-        """
-        self._gym_env.render()
+    def init_plot(self, p_figure: Figure = None, p_plot_settings: list = ..., p_plot_depth: int = 0, p_detail_level: int = 0, p_step_rate: int = 0, **p_kwargs):
+        if self._visualize: self._gym_env.render()
 
 
 ## -------------------------------------------------------------------------------------------------
-    def update_plot(self):
+    def update_plot(self, **p_kwargs):
         """
         Updating the actual plot, deployed by render functionality from OpenAI Gym.
 
         """
-        self._gym_env.render()
+        if self._visualize: self._gym_env.render()
 
 
 ## -------------------------------------------------------------------------------------------------
@@ -399,8 +444,9 @@ class WrEnvMLPro2GYM(Wrapper, gym.Env):
             Log level (see constants of class Log). Default = Log.C_LOG_ALL.
     """
 
-    C_TYPE              = 'Wrapper MLPro -> OpenAI Gym'
+    C_TYPE              = 'Wrapper MLPro2Gym'
     C_WRAPPED_PACKAGE   = 'gym'
+    C_MINIMUM_VERSION   = '0.21.0'
     metadata            = {'render.modes': ['human']}
 
 ## -------------------------------------------------------------------------------------------------
@@ -433,6 +479,13 @@ class WrEnvMLPro2GYM(Wrapper, gym.Env):
 
         self.new_step_api = p_new_step_api
 
+        if self.installed_version < "0.25.0":
+            self.reset = self.reset_old
+            if self.new_step_api:
+                raise Error("Could not use new_step_api on current Gym Version")
+        else:
+            self.reset = self.reset_new
+
         self.first_refresh = True
     
         self.log(self.C_LOG_TYPE_I, 'MLPro Environment has been sucessfully wrapped as Gym Environment.')
@@ -459,8 +512,9 @@ class WrEnvMLPro2GYM(Wrapper, gym.Env):
         space = None
         action_dim = p_mlpro_space.get_num_dim()
         id_dim = p_mlpro_space.get_dim_ids()[0]
+        base_set = p_mlpro_space.get_dim(id_dim).get_base_set()
         if len(p_mlpro_space.get_dim(id_dim).get_boundaries()) == 1:
-            space = gym.spaces.Discrete(p_mlpro_space.get_dim(id_dim).get_boundaries()[0])
+            space = gym.spaces.Discrete(p_mlpro_space.get_dim(id_dim).get_boundaries()[0]+1)
         else:
             lows = []
             highs = []
@@ -547,11 +601,11 @@ class WrEnvMLPro2GYM(Wrapper, gym.Env):
             info["TimeLimit.truncated"] = state.get_timeout()
             return obs, reward.get_overall_reward(), terminated, info
 
-    
-## -------------------------------------------------------------------------------------------------
-    def reset(self, seed=None, return_info=False, options=None):
+
+    def reset_new(self, seed=None, return_info=False, options=None):
         """
         Resets the environment to an initial state and returns the initial observation.
+        This is for new gym version.
 
         Parameters
         ----------
@@ -595,12 +649,63 @@ class WrEnvMLPro2GYM(Wrapper, gym.Env):
         else:
             obs = np.array(self._mlpro_env.get_state().get_values())
         
+        info = {}
         if return_info:
-            info = {}
             return obs, info
         else:
             return obs
 
+
+## -------------------------------------------------------------------------------------------------
+    def reset_old(self):
+        """
+        Resets the environment to an initial state and returns the initial observation.
+        This is for old gym version.
+
+        Parameters
+        ----------
+        seed : int, optional
+           The seed that is used to initialize the environment's PRNG.
+           If the environment does not already have a PRNG and ``seed=None`` (the default option) is passed,
+           a seed will be chosen from some source of entropy (e.g. timestamp or /dev/urandom).
+           However, if the environment already has a PRNG and ``seed=None`` is passed, the PRNG will *not* be reset.
+           If you pass an integer, the PRNG will be reset even if it already exists.
+           Usually, you want to pass an integer *right after the environment has been initialized and then never again*.
+           Please refer to the minimal example above to see this paradigm in action.
+           The default is None.
+        return_info : bool
+            If true, return additional information along with initial observation.
+            This info should be analogous to the info returned in :meth:`step`.
+            The default is False.
+        options : dict, optional
+            Additional information to specify how the environment is reset (optional,
+            depending on the specific environment). The default is None.
+
+        Returns
+        -------
+        obs : object
+            This will be an element of the environment's :attr:`observation_space`.
+            This may, for instance, be a numpy array containing the positions and velocities of certain objects.
+        info : dict
+            It contains auxiliary diagnostic information (helpful for debugging, learning, and logging).
+            This might, for instance, contain: metrics that describe the agent's performance state, variables that are
+            hidden from observations, or individual reward terms that are combined to produce the total reward.
+            It also can contain information that distinguishes truncation and termination, however this is deprecated in favour
+            of returning two booleans, and will be removed in a future version.
+
+        """
+        # We need the following line to seed self.np_random
+        
+        self._mlpro_env.reset()
+        obs = None
+        if isinstance(self.observation_space, gym.spaces.Box):
+            obs = np.array(self._mlpro_env.get_state().get_values(), dtype=np.float32)
+        else:
+            obs = np.array(self._mlpro_env.get_state().get_values())
+        
+        info = {}
+        return obs
+
     
 ## -------------------------------------------------------------------------------------------------
     def render(self, mode='human'):
